#include "mpff.h"
#include "xmpn.h"
#include <stdio.h>
 
/*
@function: mpff_field_init
@section: mpff/init
@proto: void mpff_field_init(mpff_field ff, mpz_t p, mpz_t pol)
@desc: Initialize ff as the finite field $\mathbb{F}_p[X]/(P)$.
Currently pol must be equal to $P(2^e)$ and P has coefficients between 
$0$ and $2^e-1$ where \verb!e=mpz_size(p)*mp_bits_per_limb!.
The current implementation require $P$ to be monic.\\
Use mpff\_field\_init\_unpack instead.
*/

void
mpff_field_init(mpff_field ff, mpz_t p, mpz_t pol)
{
  mpffn_field_init(ff, p->_mp_d, p->_mp_size, pol->_mp_d, pol->_mp_size,
      (void *)0L,(void *)0L);
}

/*
@function: mpff_field_init_unpack
@section: mpff/init
@proto: void mpff_field_init_unpack(mpff_field ff, mpz_t p, mpz_t pol, mp_size_t e)
@desc: Initialize ff as the finite field $\mathbb{F}_p[X]/(P)$  where $P$ is
defined by $P(2^e)=x$ and P has coefficients between $0$ and $2^e-1$.
The current implementation require $P$ to be monic.
*/

void
mpff_field_init_unpack(mpff_field ff, mpz_t p, mpz_t pol, mp_size_t packet_size)
{
  mp_size_t bits=mpz_sizeinbase(pol,2);
  mp_size_t packet_number=(bits+packet_size-1)/packet_size;
  mpffn_field_init_unpack(ff, p->_mp_d, p->_mp_size, pol->_mp_d,
      packet_number, packet_size,
      (void *)0L,(void *)0L);
}

/*
@function: mpff_field_clear
@section: mpff/init
@proto: void mpff_field_clear(mpff_field ff)
@desc: Free the space occupied by \ff.
This invalidate all `mpff\_t' variables that reference \ff.
*/

void
mpff_field_clear(mpff_field ff)
{
  mpffn_field_clear(ff);
}

/*
@function: mpff_init
@section: mpff/init
@proto: void mpff_init(mpff_t r, mpff_field ff)
@desc: Initialize $r$ as an element of the finite field \ff,
and set its value to $0$.
*/

void
mpff_init(mpff_t r, mpff_field ff)
{
  r->ff=ff;
  r->d=ff->alloc(ff->size_elt);
  mpffn_zero(r->d,ff);
}

/*
@function: mpff_clear
@section: mpff/init
@proto: void mpff_clear(mpff_t r)
@desc: Free the space occupied by r.
Call this function for all `mpff\_t' variables when you are done with them.
*/

void
mpff_clear(mpff_t r)
{
  r->ff->free(r->d, r->ff->size_elt);
}

/*
@function: mpff_set
@section: mpff/assign
@proto: void mpff_set(mpff_t r, mpff_t x)
@desc: Set $r$ to $x$. 
*/

void
mpff_set(mpff_t r, mpff_t x)
{
  mpffn_copy(r->d, x->d, r->ff);
}

/*
@function: mpff_set_z
@section: mpff/assign
@proto: void mpff_set_z(mpff_t r, mpz_t x)
@desc: Set $r$ to the finite field element defined by P.
Currently x must be equal to $P(2^e)$ where 
\verb!e=mpz_size(p)*mp_bits_per_limb!.
*/

void
mpff_set_z(mpff_t r, mpz_t x)
{
  mp_size_t s=x->_mp_size;
  mp_size_t t=r->ff->size_elt;
  
  if ( s > t )
    mpffn_error("Integer too large in mpff_set_z");
  if ( s < t)
    xmpn_zero(r->d, t);
  xmpn_copy(r->d, x->_mp_d, s);
}

/*
@function: mpff_set_ui
@section: mpff/assign
@proto: void mpff_set_ui(mpff_t r, unsigned long int x)
@desc: Set $r$ to $x$. 
*/

void
mpff_set_ui(mpff_t r, unsigned long int x)
{
  mpffn_zero(r->d, r->ff);
  r->d[0]=x;
  /*mpffn_pol_canon(r, r, 1, ff);*/
}

/*
@function: mpff_init_set_z
@section: mpff/initassign
@proto: void mpff_init_set_z(mpff_t r, mpff_field ff, mpz_t x)
@desc: Initialize $r$ as an element of the finite field \ff,
and set its value to $x$.
*/

void
mpff_init_set_z(mpff_t r, mpff_field ff, mpz_t x)
{
  mpff_init(r,ff);
  mpff_set_z(r,x);
}

/*
@function: mpff_init_set_ui
@section: mpff/initassign
@proto: void mpff_init_set_ui(mpff_t r, mpff_field ff, unsigned long int x)
@desc: Initialize $r$ as an element of the finite field \ff,
and set $r$ to the finite field element defined by P.
Currently x must be equal to $P(2^e)$ where 
\verb!e=mpz_size(p)*mp_bits_per_limb!.
*/

void
mpff_init_set_ui(mpff_t r, mpff_field ff, unsigned long int x)
{
  mpff_init(r,ff);
  mpff_set_ui(r,x);
}

/*
@function: mpff_unpack_z
@section: mpff/assign
@proto: void mpff_unpack_z(mpff_t r, mpz_t x, mp_size_t e)
@desc: Set $r$ to the finite field element defined by P, where P
is defined by $P(2^e)=x$ and P has coefficients between $0$ and $2^e-1$.
*/

void
mpff_unpack_z(mpff_t r, mpz_t x, mp_size_t packet_size)
{
  mp_size_t bits=mpz_sizeinbase(x,2);
  mp_size_t packet_number=(bits+packet_size-1)/packet_size;
  if (packet_number>r->ff->lt)
    mpffn_error("too many packets in mpff_unpack_z");
  mpffn_unpack(r->d, x->_mp_d, packet_size, packet_number, r->ff);
}

/*
@function: mpff_init_unpack_z
@section: mpff/initassign
@proto: void mpff_init_unpack_z(mpff_t r, mpff_field ff, mpz_t x, mp_size_t e)
@desc: Initialize $r$ as an element of the finite field \ff,
and set $r$ to the finite field element defined by P, where P
is defined by $P(2^e)=x$ and P has coefficients between $0$ and $2^e-1$.
*/

void
mpff_init_unpack_z(mpff_t r, mpff_field ff, mpz_t x, mp_size_t packet_size)
{
  mpff_init(r,ff);
  mpff_unpack_z(r,x,packet_size);
}

/*
@function: mpff_add
@section: mpff/arith
@proto: void mpff_add(mpff_t r, mpff_t x, mpff_t y)
@desc: Set $r$ to $x+y$. 
*/

void mpff_add(mpff_t r, mpff_t x, mpff_t y)
{
  mpffn_add(r->d, x->d, y->d, r->ff);
}

/*
@function: mpff_sub
@section: mpff/arith
@proto: void mpff_sub(mpff_t r, mpff_t x, mpff_t y)
@desc: Set $r$ to $x-y$. 
*/


void mpff_sub(mpff_t r, mpff_t x, mpff_t y)
{
  mpffn_sub(r->d, x->d, y->d, r->ff);
}

/*
@function: mpff_add_ui
@section: mpff/arith
@proto: void mpff_add_ui(mpff_t r, mpff_t x, unsigned long int y)
@desc: Set $r$ to $x+y$. 
*/

void mpff_add_ui(mpff_t r, mpff_t x, unsigned long int y)
{
  mpffn_add_ui(r->d, x->d, y, r->ff);
}

/*
@function: mpff_sub_ui
@section: mpff/arith
@proto: void mpff_sub_ui(mpff_t r, mpff_t x, unsigned long int y)
@desc: Set $r$ to $x-y$. 
*/

void mpff_sub_ui(mpff_t r, mpff_t x, unsigned long int y)
{
  mpffn_sub_ui(r->d, x->d, y, r->ff);
}

/*
@function: mpff_neg
@section: mpff/arith
@proto: void mpff_neg(mpff_t r, mpff_t x)
@desc: Set $r$ to $-x$.
*/

void mpff_neg(mpff_t r, mpff_t x)
{
  mpffn_neg(r->d, x->d, r->ff);
}

/*
@function: mpff_mul
@section: mpff/arith
@proto: void mpff_mul(mpff_t r, mpff_t x, mpff_t y)
@desc: Set $r$ to $xy$.
*/

void mpff_mul(mpff_t r, mpff_t x, mpff_t y)
{
  mpffn_mul(r->d, x->d, y->d, r->ff);
}

/*
@function: mpff_mul_ui
@section: mpff/arith
@proto: void mpff_mul_ui(mpff_t r, mpff_t x, unsigned long int y)
@desc: Set $r$ to $xy$.
*/


void mpff_mul_ui(mpff_t r, mpff_t x, unsigned long int y)
{
  if (r->ff->size_p==1)
    y %= r->ff->p[0];
  mpffn_mul_ui(r->d, x->d, y, r->ff);
}

/*
@function: mpff_mul_z
@section: mpff/arith
@proto: void mpff_mul_z(mpff_t r, mpff_t x, mpz_t y)
@desc: Set $r$ to $xy$.
*/


void mpff_mul_z(mpff_t r, mpff_t x, mpz_t y)
{
  mpffn_mul_z(r->d, x->d, y, r->ff);
}

/*
@function: mpff_inv
@section: mpff/arith
@proto: void mpff_inv(mpff_t r, mpff_t y)
@desc: set $r$ to $x^{-1}$.
*/

void mpff_inv(mpff_t r, mpff_t x)
{
  mpffn_inv(r->d, x->d, r->ff);
}

/*
@function: mpff_div
@section: mpff/arith
@proto: void mpff_div(mpff_t r, mpff_t x, mpff_t y)
@desc: set $r$ to $x/y$.
*/

void mpff_div(mpff_t r, mpff_t x, mpff_t y)
{
  mpffn_div(r->d, x->d, y->d, r->ff);
}

/*
@function: mpff_pow_ui
@section: mpff/arith
@proto: void mpff_pow_ui(mpff_t r, mpff_t x, unsigned long int y)
@desc: Set $r$ to $x^y$.
*/

void mpff_pow_ui(mpff_t r, mpff_t x, unsigned long int y)
{
  mpffn_pow_ui(r->d, x->d, y, r->ff);
}

/*
@function: mpff_cmp
@section: mpff/cmp
@proto: int mpff_cmp(mpff_t x, mpff_t y)
@desc: Compare $x$ and $y$.  Return $0$ if $x=y$, or $1$ if $x \neq y$.
*/

int mpff_cmp(mpff_t x, mpff_t y)
{
  return mpffn_cmp(x->d, y->d, x->ff);
}

/*
@function: mpff_sgn
@section: mpff/cmp
@proto: int mpff_sgn(mpff_t x)
@desc: Return $1$ if $x\neq 0$, $0$ if $x=0$.
*/

int mpff_sgn(mpff_t x)
{
  return mpffn_sgn(x->d, x->ff);
}

/*
@function: mpff_out_str
@section: mpff/io
@proto: size_t mpff_out_str(FILE *STREAM, int BASE, mpff_t x)
@desc: Output x on stdio stream STREAM, in GP format.  This function is currently only for debugging purpose.  Return the number of bytes written, or if an error occurred, return $0$.
*/

size_t
mpff_out_str(FILE *fout, int base, mpff_t x)
{
  return mpffn_out_str(fout,base,x->d,x->ff);
}

