#include "mpffn.h"
#include "mpffn_priv.h"
#include "xmpn.h"
#include <stdlib.h>

static mp_ptr
mpffn_alloc(mp_size_t l)
{
  mp_ptr r = (mp_ptr) malloc(l*sizeof(mp_limb_t));
  if (!r) mpffn_error("malloc failure");
  return r;
}

static void
mpffn_free(mp_ptr r, mp_size_t l)
{
  free(r);
}
/*
 * We must be called *after* mpffn_field_alloc_tmp()
 * since we need tmp space.
@function: mpffn_field_init_impl
@section: none
@temp: init_impl=ff->lp
*/

static void
mpffn_field_init_impl(mpffn_field *ff)
{
  mp_srcptr p;
  mp_ptr pol=ff->pol;
  mp_ptr z=ff->tmp->init_impl;
  long lt=ff->lt, lp=ff->lp;
  long le=ff->size_elt;
  long leader=le-lp;
  long i;
  if (!xmpn_is_one(ff->pol+ff->size_elt,ff->size_p))
  {
    /* Polynomial is not monic. We fix this here. */
    mpffn_invmod(z,pol+le,ff);
    mpffn_mul_mpn(pol,pol,z,ff);
    xmpn_zero(pol+le,lp);
    ff->pol[le]=1;
  }
  ff->impl->leader=leader;
  for(p=ff->pol+leader, i=lt; i>0; i--, p-=lp)
    if (!xmpn_is_zero(p,lp))
      break;
  ff->impl->l_pol=i;
  ff->impl->mg=ff->alloc(le*2);
  mpffn_pol_montgomery(ff->impl->mg,ff->pol,lt+1,ff);
}

static void
mpffn_field_free_impl(mpffn_field *ff)
{
  ff->free(ff->impl->mg,ff->size_elt*2);
}

/*
@function: mpffn_field_init
@section: mpffn/init
@proto: void mpffn_field_init(mpffn_field *ff, mp_ptr p, mp_size_t size_p, 
        mp_srcptr pol, mp_size_t size_pol, 
        mp_ptr (*ff_alloc)(mp_size_t l),
        void (*ff_free)(mp_ptr r,mp_size_t l))
@desc: Initialize the data structure \ff\ to the finite field of characteristic
$\{p,size\_p\}$ defined by the polynomial $\{pol,size\_pol\}$.
You can pass NULL for ff\_alloc() and ff\_free() to use the default memory
allocation functions. Else
\begin{itemize}
\item r=ff\_alloc(l) must allocate a bloc r of l mp\_limb\_t.
\item ff\_free(r,l) must free a bloc r of l mp\_limb\_t previously 
allocated by r=ff\_alloc(l).
\end{itemize}
*/

void
mpffn_field_init(mpffn_field *ff, mp_ptr p, mp_size_t size_p, mp_srcptr pol, mp_size_t size_pol, mp_ptr (*ff_alloc)(mp_size_t l), void (*ff_free)(mp_ptr r,mp_size_t l))
{
  long lp=size_p;
  long lt=size_pol-1;
  ff->alloc=ff_alloc?ff_alloc:mpffn_alloc;
  ff->free=ff_free?ff_free:mpffn_free;
  ff->size_p=size_p;
  ff->size_pol=size_pol;
  ff->lp=lp;
  ff->lt=lt;
  ff->size_elt=lp*lt;
  ff->size_sqr=2*lp*lt;
  ff->p=ff->alloc(lp);
  xmpn_zero(ff->p,lp);
  xmpn_copy(ff->p,p,size_p);
  ff->pol=ff->alloc(lp*(lt+1));
  xmpn_zero(ff->pol,lp*(lt+1));
  xmpn_copy(ff->pol,pol,ff->size_pol);
  ff->tmp = (struct mpffn_field_tmp_s *) ff->alloc
    (sizeof(*ff->tmp)/sizeof(mp_limb_t));
  ff->impl=(struct mpffn_field_impl_s *) ff->alloc
    (sizeof(*ff->impl)/sizeof(mp_limb_t));
  mpffn_field_alloc_tmp(ff);
  mpffn_field_init_impl(ff);
}

/*
@function: mpffn_field_init_unpack
@section: mpffn/init
@proto: void mpffn_field_init_unpack(mpffn_field *ff, mp_ptr p, mp_size_t
size_p, mp_srcptr pol, mp_size_t packet_number, mp_size_t packet_size, mp_ptr
(*ff_alloc)(mp_size_t l), void (*ff_free)(mp_ptr r,mp_size_t l))
@desc: Initialize the data structure \ff\ to the finite field of characteristic
$\{p,size\_p\}$ defined by the polynomial $\{pol,packet\_number,packet\_size\}$.
You can pass NULL for ff\_alloc() and ff\_free() to use the default memory
allocation functions. Else
\begin{itemize}
\item r=ff\_alloc(l) must allocate a bloc r of l mp\_limb\_t.
\item ff\_free(r,l) must free a bloc r of l mp\_limb\_t previously 
allocated by r=ff\_alloc(l).
\end{itemize}
*/

void
mpffn_field_init_unpack(mpffn_field *ff, mp_ptr p, mp_size_t size_p, mp_srcptr pol, mp_size_t packet_number, mp_size_t packet_size, mp_ptr (*ff_alloc)(mp_size_t l), void (*ff_free)(mp_ptr r,mp_size_t l))
{
  long lp=size_p;
  long size_pol=packet_number*size_p;
  long lt=size_pol-1;
  ff->alloc=ff_alloc?ff_alloc:mpffn_alloc;
  ff->free=ff_free?ff_free:mpffn_free;
  ff->size_p=size_p;
  ff->size_pol=size_pol;
  ff->lp=lp;
  ff->lt=lt;
  ff->size_elt=lp*lt;
  ff->size_sqr=2*lp*lt;
  ff->p=ff->alloc(lp);
  xmpn_zero(ff->p,lp);
  xmpn_copy(ff->p,p,size_p);
  ff->pol=ff->alloc(lp*(lt+1));
  xmpn_zero(ff->pol,lp*(lt+1));
  mpffn_unpack(ff->pol,pol,packet_size,packet_number,ff);
  ff->tmp = (struct mpffn_field_tmp_s *) ff->alloc
    (sizeof(*ff->tmp)/sizeof(mp_limb_t));
  ff->impl=(struct mpffn_field_impl_s *) ff->alloc
    (sizeof(*ff->impl)/sizeof(mp_limb_t));
  mpffn_field_alloc_tmp(ff);
  mpffn_field_init_impl(ff);
}


/*
@function: mpffn_field_clear
@section: mpffn/init
@proto: void mpffn_field_clear(mpffn_field *ff)
@desc: Free the data structure \ff\ .
*/
  
void
mpffn_field_clear(mpffn_field *ff)
{
  long lp=ff->size_p;
  long lt=ff->size_pol-1;
  mpffn_field_free_impl(ff);
  mpffn_field_free_tmp(ff);
  ff->free((mp_ptr)ff->impl,sizeof(*ff->impl)/sizeof(mp_limb_t));
  ff->free((mp_ptr)ff->tmp,sizeof(*ff->tmp)/sizeof(mp_limb_t));
  ff->free(ff->pol,ff->lp*(lt+1));
  ff->free(ff->p,lp);
}
